@@ -16,6 +16,7 @@ __PACKAGE__->config(
 
     # simple tag helpers
     tag => \&tag,
+    div_tag => \&div_tag,
     form_tag => \&form_tag,
     label_tag => \&label_tag,
     input_tag => \&input_tag,
@@ -25,11 +26,16 @@ __PACKAGE__->config(
     options_for_select => \&options_for_select,
     checkboxes_from_collection => \&checkboxes_from_collection,
     model_errors_for => \&model_errors_for,
+    namespace => \&namespace,
+    namespace_id_with => \&namespace_id_with,
+    namespace_name_with => \&namespace_name_with,
+
 
     # tag helpers with an underlying model
     form_for => \&form_for,
     label => \&label,
     input => \&input,
+    hidden => \&hidden,
     errors_for => \&errors_for,
     model_errors => \&model_errors,
     human_model_name => \&human_model_name,
@@ -40,28 +46,10 @@ __PACKAGE__->config(
   },
 );
 
-sub errors_box {
-  my ($self, $c, $model, %attrs) = @_;
-  my @errors = ();
-  if(blessed $model) {
-    @errors = $model->errors->full_messages;
-  } elsif($model) {
-    @errors = ($model);
-  }
-  if(@errors) {
-    my $max_errors = $attrs{max_errors} ? delete($attrs{max_errors}) : scalar(@errors);
-    my $errors = join '', map { "<li>$_" } @errors[0..($max_errors-1)];
-    my $attrs =  join ' ', map { "$_='$attrs{$_}'"} keys %attrs;
-    return b("<div $attrs/>$errors</div>");
-  } else {
-    return '';
-  } 
-}
-
 sub _stringify_attrs {
   my %attrs = @_;
   return unless %attrs;
-  return my $attrs =  join ' ', map { "$_='@{[ $attrs{$_}||'' ]}'"} keys %attrs;
+  return my $attrs =  join ' ', map { "$_='@{[ defined($attrs{$_}) ? $attrs{$_} : '' ]}'"} keys %attrs;
 }
 
 sub _parse_proto {
@@ -94,7 +82,6 @@ sub style_attr {
   return $self->attr($c, 'style', $attr, @_);
 }
 
-
 sub tag {
   my ($self, $c, $name, @proto) = @_;
   my ($content, %attrs) = _parse_proto(@proto);
@@ -110,6 +97,11 @@ sub tag {
   return b $tag;
 }
 
+sub div_tag {
+  my ($self, $c, @proto) = @_;
+  return $self->tag($c, 'div', @proto);
+}
+
 sub form_tag {
   my ($self, $c, @proto) = @_;
   return $self->tag($c, 'form', @proto);
@@ -175,7 +167,7 @@ sub options_from_collection_for_select {
 
 # %= checkboxes_from_collection 'person_roles.role', $roles, +{value_field=>'id', label_field=>'name', ... }
 sub checkboxes_from_collection {
-  my ($self, $c, $field_proto, $collection, @proto) = @_;
+  my ($self, $c, $field_proto, $bridge, $collection, @proto) = @_;
   my ($content, %attrs) = _parse_proto(@proto);
   my $model = $c->stash->{'valiant.view.form.model'};
   my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
@@ -183,39 +175,41 @@ sub checkboxes_from_collection {
   my $value = exists($attrs{value_field}) ? delete($attrs{value_field}) : 'id';
   my $label = exists($attrs{label_field}) ? delete($attrs{label_field}) : 'label';
 
-  my $key = '';
-  my $field_model = $model;
-  my @field_path = split('\.', $field_proto);
-  foreach my $part (@field_path) {
-    my $info = $field_model->result_source->relationship_info($part);
-
-    if($info) {
-      ($key) = keys %{$info->{attrs}{fk_columns}};  # only works with a single field PK for now
-      $field_model = $field_model->related_resultset($part);
-    } elsif($info = $field_model->_m2m_metadata->{$part}) {
-      my $rs_method = $info->{rs_method};
-      $field_model = $field_model->$rs_method;
-      ($key) = $field_model->result_source->primary_columns;  # only works with a single field FK for now
-    }
-  }
+  my $field_model_rs = $model->$field_proto;
+  my ($key, $target) = %$bridge;
 
   my $idx = 0;
   my @tags = ();
-  my $related_part = shift @field_path;
+  my @each = $field_model_rs->all;
+  my @primary_columns = $field_model_rs->result_source->primary_columns;
+
   foreach my $item($collection->all) {
-    local $c->stash->{'valiant.view.form.namespace'} = [@namespace, $related_part, $idx];
+    local $c->stash->{'valiant.view.form.namespace'} = [@namespace, $field_proto, $idx];
     local $c->stash->{'valiant.view.form.model'} = $item;
 
-    my %checked = $field_model->contains($item) ? (checked=>1) : ();
+    my ($checked) = grep { $_ }
+      map { $_->$key eq $item->$target ? $_ : undef } 
+      grep { !$_->is_removed }
+      @each;
+    my %checked = $checked ? (checked=>1) : ();
 
     push @tags, b "<div class='form-check'>";
-    push @tags, $self->input($c, $key, +{type=>'checkbox', value=>$item->$value, class=>'form-check-input', %checked});
-    push @tags, $self->label($c, $key, +{class=>'form-check-label'}, sub { $item->$label } );
+
+    if($checked && $checked->in_storage) {
+      foreach my $primary_column (@primary_columns) {
+        push @tags, $self->hidden($c, "$primary_column", +{ value=> $checked->get_column($primary_column) });
+      }
+    } else {
+      push @tags, $self->hidden($c, "$key", +{value=>$item->$value});
+    }
+
+    push @tags, $self->hidden($c, "_delete", +{value=>'1', id=>$self->namespace_id_with($c, 'hidden') });
+    push @tags, $self->input($c, '_delete', +{type=>'checkbox', value=>'0', class=>'form-check-input', %checked});
+    push @tags, $self->label($c, '_delete', +{class=>'form-check-label'}, sub { $item->$label } );
     push @tags, b "</div>";
 
     $idx++;
   }
-
   return b @tags;
 }
 
@@ -283,9 +277,9 @@ sub label {
   my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
 
   $attrs{for} ||= join '_', (@namespace, $field);
-  $content ||= sub { $model->human_attribute_name($field) };
-  
-  return $self->label_tag($c, \%attrs, $content);
+  my $inner_content = $content ? $content->($field, $model, \%attrs) : $model->human_attribute_name($field);
+
+  return $self->label_tag($c, \%attrs, $inner_content);
 }
 
 sub input {
@@ -295,11 +289,18 @@ sub input {
   my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
   my @errors = $model->errors->messages_for($field);
 
+  my @errors_classes = ();
+  if(my $errors_classes_proto = delete($attrs{errors_classes})) {
+    push @errors_classes, ref($errors_classes_proto) ? @$errors_classes_proto : ($errors_classes_proto);
+  }
+
   $attrs{type} ||= 'text';
   $attrs{id} ||= join '_', (@namespace, $field);
   $attrs{name} ||= join '.', (@namespace, $field);
   $attrs{value} = ($model->read_attribute_for_validation($field) || '') unless defined($attrs{value});
-  $attrs{class} .= ' is-invalid' if @errors;
+  $attrs{class} .= " @{[ join ' ', @errors_classes ]}" if @errors;
+
+  delete $attrs{checked} if exists($attrs{checked}) && !$attrs{checked};
 
   return $self->input_tag($c, \%attrs, $content);
 }
@@ -319,17 +320,21 @@ sub errors_for {
 
   return '' unless @errors;
 
-  my $class = $attrs{class}||'';
-  $class .= ' invalid-feedback';
-  $attrs{class} = $class;
-
   my $max_errors = $attrs{max_errors} ? delete($attrs{max_errors}) : scalar(@errors);
-  my $divider = $max_errors > 1 ? '<li>' : '';
-  my $errors = join '', map { "${divider}$_" } @errors[0..($max_errors-1)];
+  @errors =  @errors[0..($max_errors-1)];
 
-  return $self->tag($c, 'div', \%attrs, $errors);
+  return $content->(\%attrs, @errors) if $content;
+  return $c->stash->{'view.content'}->{errors_for_response}->(\%attrs, @errors) if $c->stash->{'view.content'}->{errors_for_response};
+  return $self->errors_for_response($c, \%attrs, @errors);
 }
 
+  sub errors_for_response {
+    my ($self, $c, $attrs, @errors) = @_;
+    my $divider = scalar(@errors) > 1 ? '<li>' : '';
+    my $errors = join '', map { "${divider}$_" } @errors;
+    return $self->tag($c, 'div', $attrs, $errors);
+  }
+
 sub model_errors_for {
    my ($self, $c, $attribute, %attrs) = @_;
    my $model = $c->stash->{'valiant.view.form.model'};
@@ -363,6 +368,39 @@ sub model_errors {
   return $self->tag($c, 'div', \%attrs, $errors);
 }
 
+sub fields_for {
+  my ($self, $c, $related, @proto) = @_;
+  my ($content, %attrs) = _parse_proto(@proto);
+  my $model = $c->stash->{'valiant.view.form.model'};
+  my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
+
+  die "No relation '$related' for model $model" unless $model->has_relationship($related);
+
+  my $idx = 0;
+  my $content_expanded = '';
+  my $resultset = $model->related_resultset($related);
+  my $namespace = $model->relationship_info($related)->{attrs}{accessor} eq 'single' ?
+    sub { [@namespace, $related] }
+    : sub { [@namespace, $related, $idx] };
+
+  while(my $result = $resultset->next) {
+    local $c->stash->{'valiant.view.form.model'} = $result;
+    local $c->stash->{'valiant.view.form.namespace'} = $namespace->();
+
+    $content_expanded .= $content->();
+
+    if($result->in_storage) {
+      my @primary_columns = $result->result_source->primary_columns;
+      foreach my $primary_column (@primary_columns) {
+        next unless my $value = $result->get_column($primary_column);
+        $content_expanded .= $self->hidden($c, $primary_column);
+      }
+    }
+    $idx++
+  }
+  return b $content_expanded;
+}
+
 sub human_model_name {
   my ($self, $c, @proto) = @_;
   return $c->stash->{'valiant.view.form.model'}->model_name->human;
@@ -373,144 +411,40 @@ sub model {
   return  $c->stash->{'valiant.view.form.model'};
 }
 
-sub fields_for {
-  my ($self, $c, $related, @proto) = @_;
-  my ($content, %attrs) = _parse_proto(@proto);
-  my $model = $c->stash->{'valiant.view.form.model'};
-  my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
-
-
-  # This next bit is DBIC specific.   It would be more ideal to isolate this
-  # behind an abtraction.
-  die "No relation '$related' for model $model" unless $model->has_relationship($related);
-
-  my $rel_data = $model->relationship_info($related);
-  my $rel_type = $rel_data->{attrs}{accessor};
-
-  if( $rel_type eq 'single') {
-    my $result = $model->related_resultset($related)->first;
-    return '' unless $result;  # not sure this is correct
-
-    local $c->stash->{'valiant.view.form.model'} = $result;
-    local $c->stash->{'valiant.view.form.namespace'} = [@namespace, $related];
-
-    my $content_expanded = $content->();
-
-    if($result->in_storage) {
-      my @primary_columns = $result->result_source->primary_columns;
-      foreach my $primary_column (@primary_columns) {
-        next unless my $value = $result->get_column($primary_column);
-        $content_expanded .= $self->hidden($c, $primary_column);
-      }
-    }
-    return b $content_expanded;
-  }
+sub namespace {
+  my ($self, $c,) = @_;
+  return  $c->stash->{'valiant.view.form.namespace'};
 }
+
+sub namespace_name_with {
+  my ($self, $c, @with) = @_;
+  my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
+  return join '.', @namespace, @with;
+
+}
+sub namespace_id_with {
+  my ($self, $c, @with) = @_;
+  my @namespace = @{$c->stash->{'valiant.view.form.namespace'}||[]};
+  return join '_', @namespace, @with;
+}
+
+sub errors_box {
+  my ($self, $c, $model, %attrs) = @_;
+  my @errors = ();
+  if(blessed $model) {
+    @errors = $model->errors->full_messages;
+  } elsif($model) {
+    @errors = ($model);
+  }
+  if(@errors) {
+    my $max_errors = $attrs{max_errors} ? delete($attrs{max_errors}) : scalar(@errors);
+    my $errors = join '', map { "<li>$_" } @errors[0..($max_errors-1)];
+    my $attrs =  join ' ', map { "$_='$attrs{$_}'"} keys %attrs;
+    return b("<div $attrs/>$errors</div>");
+  } else {
+    return '';
+  } 
+}
+
     
 __PACKAGE__->meta->make_immutable;
-
-
-
-
-
-
-__END__
-
-
-      %= input_group 'address', begin
-        <div class='form-group'>
-          %= label current_field
-          %= input current_field, +{ class=>'form-control' }
-          %= errors_for current_field
-        </div>        
-      %= end
-
-    %= input_group 'address', +{ class=>'form-control', div_attrs=>{ class=>'form-group'}  }
-
-
-sub model_errors_for {
-   my ($self, $c, $attribute, %attrs) = @_;
-   my $model = $c->stash->{'valiant.view.form.model'};
-
-   if(my @errors = $model->errors->full_messages_for($attribute)) {
-     my $max_errors = $attrs{max_errors} ? delete($attrs{max_errors}) : scalar(@errors);
-     my $errors = join ', ', @errors[0..($max_errors-1)];
-     my $attrs =  join ' ', map { "$_='$attrs{$_}'"} keys %attrs;
-     return b("<div $attrs/>$errors</div>");
-   } else {
-     return '';
-   }
-}
-sub form_tag {
-  my ($self, $c, $url, @proto) = @_;
-  my ($content, %attrs) = _parse_proto(@proto);
-  
-  # $attr{action} ||= $c->uri_for()...
-
-  my $out = $self->tag($c, 'form', \%attrs);
-  if($content) {
-    $out .= $content->();
-    $out .= "</form>";
-    $out = b $out;
-  }
-
-  return $out;
-}
-
-
-
-
-
-
-
-
-sub form_for {
-  my ($self, $c, $model, @proto) = @_;
-  my ($content, %attrs) = _parse_proto(@proto);
-
-  if($model->can('in_storage') && $model->in_storage) {
-    my $value = $model->model_name->param_key . '_edit';
-    $attrs{id} ||= $value;
-    $attrs{class} ||= $value;
-  } else {
-    my $value = $model->model_name->param_key . '_new';
-    $attrs{id} ||= $value;
-    $attrs{class} ||= $value;
-  }
-  $attrs{method} ||= 'POST';
-
-  return $self->form_tag($c, '', %attrs, $content);
-}
-
-
-sub fields_for {
-  my ($self, $c, $model, @proto) = @_;
-  my ($content, %attrs) = _parse_proto(@proto);
-
-  local $c->stash->{'valiant.view.form.model'} = $model;
-  local $c->stash->{'valiant.view.form.namespace'}[0] = $model->model_name->param_key;
-
-  $attrs{id} ||= $model->model_name->param_key;
-  $attrs{method} ||= 'POST';
-
-  my $attrs = _stringify_attrs(%attrs);
-
-
-
-}
-
-
-
-
-
-sub text_field_tag {
-  my ($self, $c, $name, $value, $options) = @_;
-  $options ||= +{};
-  $options->{type} ||= 'text';
-  $options->{name} ||= $name;
-  $options->{id} ||= do { $name=~tr/\./_/; $name };
-  $options->{value} ||= $value;
-  return $self->tag($c, 'input', $options);
-}
-
-1;
